part of tecfy_core_client_package;

class EntityService {
  static List<EntityCategory>? categories;
  //static List<Entity> entities;
  static Map<String, Entity>? entities;
  static Map<String, List<dynamic>> dataCache = {};
  static int pageSize = 15;

  /// get value from map
  /// for example 'true' will be true
  /// and int string will be int number
  static dynamic getValue(Map<String, dynamic> data, value) {
    if (value == null || value.toString() == 'null') return null;
    if (value == 'MyUserId' || value == 'myUserId')
      return LocalConfigService.user?.id;
    var v = EntityService.lowerFirst(value);
    if (data.keys.contains(v)) return data[v];
    if (v == 'true') return true;
    if (v == 'false') return false;
    if (num.tryParse(value) != null) return num.parse(value);
    return value;
  }

  /// init entities from server or from local storage if offline
  static Future initEntities() async {
    try {
      var result = ConnectionService.activeConnection == null
          ? OperationResult(success: false)
          : await TecfyBasicApp.apiService.httpGetDynamic('entity/GetEntities');
      if (result.success) {
        TecfySyncService.syncNow();
        ConfigService.setValueString('Entity', jsonEncode(result.data));
      } else {
        result.data = jsonDecode(ConfigService.getValueString('Entity'));
        result.success = (result.data != null);
      }
      if (result.success) {
        print('-----------  GetEntities ------------ ${result.data.length}');

        List<Entity> entList = [];
        EntityService.entities = {};
        for (var cat in result.data) {
          var ent = Entity.fromMap(cat);
          entList.add(ent);
          EntityService.entities?[ent.dbName] = ent;
        }

        print('Entity count is : ${EntityService.entities?.keys.length}');
        for (var c in EntityService.categories ?? []) {
          c.entities = entList
              .where((e) => c.entities.any((ee) => ee.dbName == e.dbName))
              .toList();
        }
        print('categories count is ${EntityService.categories?.length}');
        DatabaseService.updateEntityTable();
      }
    } catch (err, t) {
      print('XXXXXXXXXXXXXXX GetEntities $err');
      print(t);
      throw "init entities Error: $err";
    }
  }

  /// get records from server of local db
  static Future<OperationResult<List<Map<String, dynamic>>>> getDataSearch(
      Entity entity, String searchText,
      {int pageIndex = 0,
      bool activeOnly = false,
      bool allRecords = false,
      int? currentPageSize}) async {
    currentPageSize = currentPageSize ?? pageSize;
    print('item search ');
    if (ConnectionService.isOfflineMode || entity.isLocal) {
      IEntityTransaction? service =
          TecfyCoreApp.entityServices[entity.dbName.toLowerCase()];

      var searchResult = await service?.search(searchText,
          skip: currentPageSize * pageIndex,
          limit: currentPageSize,
          activeOnly: activeOnly,
          allRecords: allRecords);

      if (searchResult?.success ?? false) return searchResult!;

      List<EntityFilter> f = [];
      for (var prop in entity.properties.where((p) => p.isSearchable)) {
        f.add(EntityFilter(
            propertyName: prop.dbName,
            operatorId: prop.valueType == ValueTypes.Number ? 'E' : 'C',
            value: searchText));
      }
      var d = (await EntityService.getData(entity,
              activeOnly: activeOnly,
              filter: f,
              pageIndex: pageIndex,
              filterOperator: 'or'))
          .data;
      if (d == null)
        return OperationResult(success: false, message: 'No Data Found!');

      return OperationResult(success: false, data: d);
    } else {
      var result = await TecfyBasicApp.apiService.httpGetDynamic(
          'entity/search/${entity.dbName}?text=$searchText&filters=${activeOnly ? Uri.encodeQueryComponent('[{"PropertyId":"Active","Value":true,"Operator":"E"}]') : null}&skip=${pageIndex * currentPageSize}&count=${allRecords ? 10000 : currentPageSize}');
      if (!result.success) {
        print(result.success);
        print(result.message);
        if (result.errorCodeString == 'ClientVersionNotSupported') {
          TecfyCoreApp.mustUpgrade = true;
          AppLocale.languageStreamController.add(LocalConfigService.language);
        }
        return OperationResult<List<Map<String, dynamic>>>(
            success: false, message: result.message);
      } else {
        print(' ${result.data?.length?.toString() ?? 'No Data!'}');
      }
      return OperationResult<List<Map<String, dynamic>>>(
          success: true, data: result.data.cast<Map<String, dynamic>>());
    }
  }

  /// get records from server of local db
  static Future<OperationResult<int>> getCount(Entity? entity,
      {List<EntityFilter>? filter,
      bool activeOnly = false,
      String? parentId,
      bool filterParentId = false,
      String filterOperator = 'and'}) async {
    try {
      if (entity == null) {
        return OperationResult(
            success: false, message: 'GetData: Entity cannot be null');
      }
      print('start get Count ${entity.dbName}');

      if (ConnectionService.isOfflineMode || entity.isLocal) {
        var where = [];
        var whereArgs = [];
        if (filterParentId && parentId != null && parentId.isNotEmpty) {
          where.add('parentId = ?');
          whereArgs.add(parentId);
        }
        var w = DatabaseService.getWhere(filter ?? []);
        where.addAll(w['where']);
        whereArgs.addAll(w['whereArgs']);
        var dbResult = await DatabaseService.getEntitiesCount(entity.dbName,
            where: where.join(' $filterOperator '), whereArgs: whereArgs);
        if (dbResult != null) {
          return OperationResult<int>(success: true, data: dbResult);
        }
        if (entity.isLocal)
          return OperationResult<int>(
              success: false, message: 'No Records in Local Database!');
      }
      var pagingQuery =
          '?isOr=${filterOperator == 'or'}&activeOnly=$activeOnly&filterParentId=$filterParentId';
      if (parentId != null) pagingQuery += '&parentId=$parentId';
      var result = await TecfyBasicApp.apiService.httpGetDynamic(
          'entity/getCount/${entity.dbName}' +
              ((filter?.length ?? 0) > 0
                  ? '$pagingQuery&filter=${jsonEncode(filter?.map<Map>((f) => f.toJson()).toList())}'
                  : pagingQuery));
      if (!result.success) {
        return OperationResult<int>(success: false, message: result.message);
      }
      return OperationResult<int>(success: true, data: result.data);
    } catch (err) {
      print('XXXXXXXXXXXXXXXXXXXXX EntotyService.getData');
      print(err);
      return OperationResult<int>(success: false, message: err.toString());
    }
  }

  /// get records from server of local db
  static Future<OperationResult<List<Map<String, dynamic>>>> getDataEx(
      String entity,
      {List<EntityFilter>? filter,
      int pageIndex = 0,
      bool activeOnly = false,
      bool allRecords = false,
      String? parentId,
      List<String>? entityToInclude,
      String? orderBy,
      bool orderByDesc = false,
      bool filterParentId = false,
      int? currentPageSize}) {
    return getData(EntityService.getEntity(entity),
        filter: filter,
        pageIndex: pageIndex,
        activeOnly: activeOnly,
        allRecords: allRecords,
        parentId: parentId,
        entityToInclude: entityToInclude,
        orderBy: orderBy,
        orderByDesc: orderByDesc,
        filterParentId: filterParentId,
        currentPageSize: currentPageSize);
  }

  /// get records from server of local db
  static Future<OperationResult<List<Map<String, dynamic>>>> getData(
      Entity? entity,
      {List<EntityFilter>? filter,
      int pageIndex = 0,
      bool activeOnly = false,
      bool allRecords = false,
      String? parentId,
      bool filterParentId = false,
      List<String>? entityToInclude,
      String? orderBy,
      bool orderByDesc = false,
      String filterOperator = 'and',
      int? currentPageSize}) async {
    try {
      currentPageSize = currentPageSize ?? pageSize;
      if (entity == null) {
        return OperationResult(
            success: false, message: 'GetData: Entity cannot be null');
      }
      print('start get data ${entity.dbName}');
      if (entity.permission.canView == false && entity.allowPermisison) {
        print('xxxx no view permissions ${entity.dbName}');
        return OperationResult(
            success: false, message: 'No Permission to view ${entity.nameAr}');
      }
      if (ConnectionService.isOfflineMode || entity.isLocal) {
        var where = [];
        var whereArgs = [];
        if (filterParentId && parentId != null && parentId.isNotEmpty) {
          where.add('parentId = ?');
          whereArgs.add(parentId);
        } else if (filterParentId) {
          where.add('parentId is null');
        }
        var w = DatabaseService.getWhere(filter ?? []);
        where.addAll(w['where']);
        whereArgs.addAll(w['whereArgs']);
        var dbResult = (await DatabaseService.getEntities(entity.dbName,
                where: where.join(' $filterOperator '),
                whereArgs: whereArgs,
                orderBy: orderBy == null
                    ? null
                    : '$orderBy ${orderByDesc ? "desc" : ""}',
                limit: allRecords ? 10000 : currentPageSize,
                offset: pageIndex * currentPageSize))
            ?.data;
        if (dbResult != null) {
          dbResult.sort((a, b) => (a['sortId'] ?? 1) - (b['sortId'] ?? 1));
          dbResult = await bindOtherEntities(entity, dbResult, entityToInclude);
          return OperationResult<List<Map<String, dynamic>>>(
              success: true, data: dbResult);
        }
        if (entity.isLocal)
          return OperationResult<List<Map<String, dynamic>>>(
              success: false, message: 'No Records in Local Database!');
      }
      var pagingQuery =
          '?isOr=${filterOperator == 'or'}&skip=${pageIndex * currentPageSize}&count=${allRecords ? 10000 : currentPageSize}&activeOnly=$activeOnly&filterParentId=$filterParentId';
      if (parentId != null) pagingQuery += '&parentId=$parentId';
      if (orderBy != null)
        pagingQuery += '&orderBy=$orderBy&orderByDesc=$orderByDesc';
      if (entityToInclude != null)
        pagingQuery += "&entityToInclude=${entityToInclude.join(',')}";
      var result = await TecfyBasicApp.apiService.httpGetDynamic(
          'entity/get/${entity.dbName}' +
              ((filter?.length ?? 0) > 0
                  ? '$pagingQuery&filter=${jsonEncode(filter?.map<Map>((f) => f.toJson()).toList())}'
                  : pagingQuery));
      if (!result.success) {
        print(result.success);
        print(result.message);
        if (result.errorCodeString == 'ClientVersionNotSupported') {
          TecfyCoreApp.mustUpgrade = true;
          AppLocale.languageStreamController.add(LocalConfigService.language);
        }
        return OperationResult<List<Map<String, dynamic>>>(
            success: false,
            message: result.message,
            lastVersion: result.lastVersion,
            minVersion: result.minVersion);
      } else {
        print('Record Count: ${result.data?.length?.toString() ?? 'No Data!'}');
      }
      // if(cache) dataCache[entity] =result.data;
      // if (pageIndex == 0)

      //   await DatabaseService.writeCollection(entity.dbNme.saveEntity(entity.dbName, result.data,
      //     overwriteLocal: entity.isLocal);
      return OperationResult<List<Map<String, dynamic>>>(
        success: true,
        data: result.data.cast<Map<String, dynamic>>(),
        lastVersion: result.lastVersion,
        minVersion: result.minVersion,
      );
    } catch (err) {
      print('XXXXXXXXXXXXXXXXXXXXX EntotyService.getData');
      print(err);
      return OperationResult<List<Map<String, dynamic>>>(
          success: false, message: err.toString());
    }
  }

  /// load the related entities when offlice
  static Future<List<Map<String, dynamic>>> bindOtherEntities(Entity entity,
      List<Map<String, dynamic>> _data, List<String>? entityToInclude) async {
    if (ConnectionService.isOnlineMode) return _data;

    for (var prop in entity.properties) {
      if ((!prop.isAdd && !prop.isEdit && !prop.isGrid) &&
          !(entityToInclude?.contains(prop.dbName) ?? false)) continue;
      if (prop.propertyType == PropertyTypes.Dropdown) {
        Map<String, dynamic> cache = {};
        var property = (prop as PropertyDropdown);
        for (var r in _data) {
          if (r[property.dbNameEntity] != null ||
              r[prop.dbName] == null ||
              property.entityName.isEmpty) continue;
          if (cache[r[prop.dbName]] == null) {
            cache[r[prop.dbName]] = (await EntityService.getOne(
                property.entityName, r[prop.dbName]));
          }
          r[property.dbNameEntity] = cache[r[prop.dbName]];
        }
      }
    }
    return _data;
  }

  /// get records from server of local db
  static Future<Map<String, dynamic>?> getOne(String entityName, dynamic id,
      {bool forView = false, bool withRelatedData = false}) async {
    if (id == null) return null;
    if (ConnectionService.isOfflineMode) {
      var dbResult = await DatabaseService.getEntityById(entityName, id);
      if (entityName.toLowerCase() == 'customer') {
        var balance = await EntityService.getOne('customerBalance', id);
        dbResult['balance'] = balance?['balance'];
      }
      if (dbResult == null || !withRelatedData) return dbResult;
      var entity = getEntity(entityName);
      for (var prop in entity?.properties ?? []) {
        if (prop.propertyType == PropertyTypes.Dropdown) {
          var p = prop.dbName.toString().substring(0, prop.dbName.length - 2);
          if (dbResult[p] != null) continue;
          dbResult[p] = (await getOne(
            (prop as PropertyDropdown).entityName,
            dbResult[prop.dbName],
            forView: false,
            withRelatedData: false,
          ));
        }
        if (prop.propertyType == PropertyTypes.OtherEntityMany) {
          dbResult[prop.dbName] = (await getDataEx(
                  (prop as PropertyOtherEntityMany).entityName,
                  allRecords: true,
                  filter: [
                EntityFilter(
                    value: id,
                    operatorId: "E",
                    propertyName: "${entity?.dbName}Id")
              ]))
              .data;
        }
      }
      return dbResult;
    } else {
      if (withRelatedData) forView = true;
      var result = await TecfyBasicApp.apiService.httpGetDynamic(
          'entity/GetById/$entityName/${Uri.encodeComponent(id)}?view=$forView');
      if (!result.success &&
          result.errorCodeString == 'ClientVersionNotSupported') {
        TecfyCoreApp.mustUpgrade = true;
        AppLocale.languageStreamController.add(LocalConfigService.language);
      }
      if (!result.success) {
        print(result.toString());
      }
      return result.data;
    }
  }

  /// delete record from server of local db
  Future<OperationResult> delete(Entity entity, dynamic id,
      {bool withTrans = false}) async {
    IEntityTransaction? service =
        TecfyCoreApp.entityServices[entity.dbName.toLowerCase()];
    OperationResult result;
    if (!withTrans) DatabaseService.trans = [];
    if (service != null) {
      var serviceResult = await service.beforeDelete(null, id);
      if (!serviceResult.success) {
        return serviceResult;
      }
    }

    if (ConnectionService.isOfflineMode || entity.isLocal) {
      result = await DatabaseService.deleteEntityById(entity.dbName, id,
          withTrans: true);
    } else {
      result = await TecfyBasicApp.apiService.httpDeleteDynamic(
          'entity/${entity.dbName}/${Uri.encodeComponent(id)}');
    }
    if (!result.success) return result;
    if (service != null) {
      var serviceResult = await service.afterDelete(null, id);
      if (!serviceResult.success) {
        return serviceResult;
      }
    }

    return OperationResult(
        success: withTrans || ConnectionService.isOnlineMode
            ? true
            : await DatabaseService.execTrans());
  }

  /// get entity object
  static Entity? getEntity(String? name) {
    if (name == null) return null;
    if (name == '_TestEntity') {
      return Entity.fromMap({
        'dbName': '_TestEntity',
        'nameAr': 'TestEntity',
        'nameEn': 'TestEntity',
        'isLocal': true,
        'permission': {
          'canView': true,
          'canUpdate': true,
          'canAdd': true,
          'canDelete': true
        },
        'properties': [
          {
            'dbName': 'name',
            'nameAr': 'Name',
            'nameEn': 'Name',
            'valueType': ValueTypes.Text.index,
            'propertyType': PropertyTypes.Text.index,
            'fullWidth': false,
            'isEdit': true,
            'isAdd': true,
            'isGrid': true,
            'isRequired': true,
            'isSearchable': true,
            'isDisplayField': true
          },
          {
            'dbName': 'description',
            'nameAr': 'description',
            'nameEn': 'description',
            'valueType': ValueTypes.Text.index,
            'propertyType': PropertyTypes.Text.index,
            'fullWidth': false,
            'isEdit': true,
            'isAdd': true,
            'isGrid': true,
            'isRequired': true,
            'isSearchable': true,
            'isDisplayField': true
          },
          {
            'dbName': 'area',
            'nameAr': 'Area',
            'nameEn': 'Area',
            'valueType': ValueTypes.Text.index,
            'propertyType': PropertyTypes.MarkImage.index,
            'fullWidth': true,
            'isEdit': true,
            'isAdd': true,
            'isGrid': false,
            'isRequired': true,
            'isSearchable': false,
            'isDisplayField': false
          }
        ]
      });
    }
    if (name == '_ServerConnections') {
      return Entity.fromMap({
        'isLocal': true,
        'nameAr': 'قنوات الاتصال',
        'nameEn': 'Server Connections',
        'dbName': '_ServerConnections',
        'permission': {
          'canView': true,
          'canUpdate': true,
          'canAdd': true,
          'canDelete': true
        },
        'properties': [
          {
            'dbName': 'name',
            'nameAr': 'الاسم',
            'nameEn': 'Name',
            'valueType': ValueTypes.Text.index,
            'propertyType': PropertyTypes.Text.index,
            'fullWidth': false,
            'isEdit': true,
            'isAdd': true,
            'isGrid': true,
            'isRequired': true,
            'isSearchable': true,
            'isDisplayField': true
          },
          {
            'dbName': 'server',
            'nameAr': 'عنوان الخادم',
            'nameEn': 'Server',
            'valueType': ValueTypes.Text.index,
            'propertyType': PropertyTypes.Text.index,
            'fullWidth': false,
            'isEdit': true,
            'isAdd': true,
            'isGrid': false,
            'isRequired': true,
            'isSearchable': false,
            'isDisplayField': false
          },
          {
            'dbName': 'sortId',
            'nameAr': 'الترتيب',
            'nameEn': 'Sort Id',
            'valueType': ValueTypes.Number.index,
            'propertyType': PropertyTypes.Number.index,
            'fullWidth': false,
            'isEdit': true,
            'isAdd': true,
            'isGrid': false,
            'isRequired': true,
            'isSearchable': false,
            'isDisplayField': false
          },
          {
            'dbName': 'version',
            'nameAr': 'الاصدار',
            'nameEn': 'Version',
            'valueType': ValueTypes.Text.index,
            'propertyType': PropertyTypes.Text.index,
            'fullWidth': false,
            'isEdit': false,
            'isAdd': true,
            'isGrid': true,
            'isRequired': false,
            'isSearchable': false,
            'isDisplayField': false
          },
          {
            'dbName': 'lastActive',
            'nameAr': 'اخر وقت للفحص',
            'nameEn': 'Last Active',
            'valueType': ValueTypes.Number.index,
            'propertyType': PropertyTypes.DateTime.index,
            'fullWidth': false,
            'isEdit': false,
            'isAdd': true,
            'isGrid': true,
            'isRequired': false,
            'isSearchable': false,
            'isDisplayField': false
          },
          {
            'dbName': 'isCloud',
            'nameAr': 'هل سحابي؟',
            'nameEn': 'Is Cloud',
            'valueType': ValueTypes.Boolean.index,
            'propertyType': PropertyTypes.Boolean.index,
            'fullWidth': false,
            'isEdit': true,
            'isAdd': true,
            'isGrid': false,
            'isRequired': true,
            'isSearchable': false,
            'isDisplayField': false
          },
          {
            'dbName': 'isHTTPS',
            'nameAr': 'HTTPS',
            'nameEn': 'HTTPS',
            'valueType': ValueTypes.Boolean.index,
            'propertyType': PropertyTypes.Boolean.index,
            'fullWidth': false,
            'isEdit': true,
            'isAdd': true,
            'isGrid': false,
            'isRequired': true,
            'isSearchable': false,
            'isDisplayField': false
          },
          {
            'dbName': 'active',
            'nameAr': 'فعال',
            'nameEn': 'Active',
            'valueType': ValueTypes.Boolean.index,
            'propertyType': PropertyTypes.Boolean.index,
            'fullWidth': false,
            'isEdit': true,
            'isAdd': true,
            'isGrid': false,
            'isRequired': true,
            'isSearchable': false,
            'isDisplayField': false
          },
          {
            'dbName': 'colorString',
            'nameAr': 'اللون',
            'nameEn': 'Color',
            'valueType': ValueTypes.Text.index,
            'propertyType': PropertyTypes.Color.index,
            'fullWidth': false,
            'isEdit': true,
            'isAdd': true,
            'isGrid': false,
            'isRequired': false,
            'isSearchable': false,
            'isDisplayField': false
          },
        ],
      });
    }
    if (entities == null) {
      print(
          'XXXXXXXXXXXXXXXXX cannot return entity [$name] becuase entities list is null');
      return null;
    }
    var e = entities?[lowerFirst(name)];
    if (e == null) {
      e = entities?[name.toLowerCase()];
    }
    if (e == null) {
      //print('XXXXXXXXXXXXXXXXX entity [$name] not found');
      return null;
    }
    return e;
  }

  /// validate entity
  static OperationResult validateEntity(Entity? entity, dynamic data) {
    print('Validate Entity : ${entity?.dbName}');
    if (entity == null) return OperationResult(success: false);
    try {
      for (var prop in entity.properties
          .where((e) => (e.isEdit || e.isAdd) && !e.readOnly)) {
        if (prop.propertyType == PropertyTypes.OtherEntityMany) {
          var subList = data[prop.dbName];
          var subEnt = getEntity((prop as PropertyOtherEntityMany).entityName);
          if (prop.isRequired && (subList == null || subList.length == 0))
            return OperationResult(
                success: false,
                message:
                    '${AppLocale.translate('Add at least one record')} (${prop.name})');
          else if (subList != null) {
            for (var sub in subList) {
              var result = validateEntity(subEnt, sub);
              if (!result.success) {
                result.message =
                    '${result.message}\r\n ${subEnt?.name} (${subList.indexOf(sub) + 1})';
                return result;
              }
            }
          }
        } else {
          var result = prop.isValid(data[prop.dbName]);
          if (!result.success) {
            print('------Error validating: ${prop.dbName}');
            print(data[prop.dbName]);
            print(result.message);
            return result;
          }
        }
      }
      return OperationResult(success: true);
    } catch (err) {
      return OperationResult(
          success: false,
          message: 'Error in Validation ${entity.dbName}: ${err.toString()}');
    }
  }

  static Future<OperationResult> _addEntity(Entity? entity, data,
      {bool localOnly = false,
      bool withTrans = false,
      bool isPosted = false,
      bool forceOnline = false,
      List<dynamic>? entitySaveConfirm}) async {
    if (entity == null)
      return OperationResult(success: false, message: 'Entity cannot be null');
    print('+++++++++++++++++++++ Start Add ${entity.dbName}');
    if (entity.dbName.startsWith('_')) {
      var localItems = await DatabaseService.readCollection(entity.dbName);
      if (!localItems.success) localItems.data = [];
      if (data['id'] == null) data['id'] = Uuid().v4().replaceAll('-', '');
      localItems.data?.add(data);
      var saveResult = await DatabaseService.writeCollection(
          entity.dbName, localItems.data ?? []);
      if (saveResult.success) ConnectionService.reloadConnections();
      return saveResult;
    }
    if (!forceOnline && (ConnectionService.isOfflineMode || entity.isLocal)) {
      try {
        if (isPosted)
          data['isPosted'] = true;
        else
          data['isPosted'] = data['isPosted'] ?? false;
      } catch (err) {}
      if (withTrans)
        DatabaseService.trans?.add(
            DbTransaction(QueryType.insert, table: entity.dbName, data: data));
      else {
        var r = await DatabaseService.insert(entity.dbName, data);
        return OperationResult(success: r > 0);
      }
      return OperationResult(success: true, data: data);
    } else {
      var entitySaveConfirmString = entitySaveConfirm == null
          ? '[]'
          : jsonEncode(entitySaveConfirm.map((e) {
              return {'dbName': e['dbName'], 'value': e['value']};
            }).toList());
      var result = await TecfyBasicApp.apiService.httpPostEx(
          'entity/${entity.dbName}?entitySaveConfirm=$entitySaveConfirmString',
          json
              .encode(data, toEncodable: customEncode)
              .replaceAll('\'', '')
              .replaceAll('\"', '\''));
      if (!result.success)
        print(result.toString());
      else
        print('${entity.dbName} added successfully');
      return result;
    }
  }

  static Future<OperationResult> _updateEntity(Entity? entity, data,
      {bool localOnly = false,
      bool withTrans = false,
      bool isPosted = false,
      bool forceOnline = false,
      List<dynamic>? entitySaveConfirm}) async {
    if (entity == null)
      return OperationResult(success: false, message: 'Entity cannot be null');
    print('+++++++++++++++++++++ Start Update ${entity.dbName}');
    if (entity.dbName.startsWith('_')) {
      var localItems = await DatabaseService.readCollection(entity.dbName);
      if (!localItems.success) return localItems;
      if (data['id'] == null) data['id'] = Uuid().v4().replaceAll('-', '');
      localItems.data?.removeWhere((element) => element['id'] == data['id']);
      localItems.data?.add(data);
      localItems.data?.sort((a, b) => a['sortId'] ?? 0 - b['sortId'] ?? 0);
      var saveResult = await DatabaseService.writeCollection(
          entity.dbName, localItems.data ?? []);
      if (saveResult.success) ConnectionService.reloadConnections();
      return saveResult;
    }

    if (!forceOnline && (ConnectionService.isOfflineMode || entity.isLocal)) {
      if (withTrans)
        DatabaseService.trans?.add(DbTransaction(QueryType.update,
            table: entity.dbName,
            data: data,
            where: 'id = ?',
            whereArgs: [data['id']]));
      else {
        var r = await DatabaseService.update(entity.dbName, data,
            where: 'id = ?', whereArgs: [data['id']]);
        return OperationResult(success: (r ?? 0) > 0);
      }
      return OperationResult(success: true);
    } else {
      var entitySaveConfirmString = entitySaveConfirm == null
          ? '[]'
          : jsonEncode(entitySaveConfirm.map((e) {
              return {'dbName': e['dbName'], 'value': e['value']};
            }).toList());
      var result = await TecfyBasicApp.apiService.httpPutEx(
          'entity/${entity.dbName}?entitySaveConfirm=$entitySaveConfirmString',
          json
              .encode(data, toEncodable: customEncode)
              .replaceAll('\'', '')
              .replaceAll('\"', '\''));
      return result;
    }
  }

  /// save record from server of local db
  static Future<OperationResult> saveEntityEx(
      String entity, dynamic data, BuildContext context, bool isAddNew,
      {bool skipValidation = false,
      bool withTrans = false,
      bool forceOnline: false,
      bool isPosted = false}) {
    return EntityService.saveEntity(
        EntityService.getEntity(entity), data, context, isAddNew,
        withTrans: withTrans, forceOnline: forceOnline, isPosted: isPosted);
  }

  /// save record from server of local db
  static Future<OperationResult> saveEntity(
      Entity? entity, dynamic data, BuildContext? context, bool isAddNew,
      {bool skipValidation = false,
      bool localOnly = false,
      bool forceOnline: false,
      bool withTrans = false,
      bool isPosted = false,
      List<dynamic>? entitySaveConfirm}) async {
    try {
      if (data == null) {
        return OperationResult(
            success: false,
            message: 'Cannot Save (${entity?.nameEn}), Data is Null');
      }
      if (data['isLocked'] == true && !isAddNew) {
        return OperationResult(success: false, message: 'Record is Locked');
      }
      if (!withTrans) DatabaseService.trans = [];
      print(
          '+++++++++++++++++++++SAVE Start [${isAddNew ? 'Add' : 'Update'}] [${entity?.dbName}] (id=${data['id']})');
      if (!skipValidation) {
        var validationResult = EntityService.validateEntity(entity, data);
        if (!validationResult.success) return validationResult;
      }

      if (data['id'] == null && ConnectionService.isOfflineMode)
        data['id'] = Uuid().v4().replaceAll('-', '');

      IEntityTransaction? service =
          TecfyCoreApp.entityServices[entity?.dbName.toLowerCase() ?? []];
      OperationResult? result;
      print('--------- Service ${service == null ? '[NOT]' : ''} Found');
      // if (forceOnline) {
      data['forceOnline'] = forceOnline;
      // }
      if (isAddNew) {
        if (service != null) {
          var serviceResult = await service.beforeAdd(context, data);
          if (!serviceResult.success) {
            return serviceResult;
          }
          if (!serviceResult.skip && serviceResult.success) {
            data = serviceResult.data;
            // result = await EntityService._addEntity(entity, serviceResult.data,
            //     localOnly: localOnly, withTrans: true);
          }
        }
        if (result == null) {
          result = await EntityService._addEntity(entity, data,
              localOnly: localOnly,
              forceOnline: forceOnline,
              withTrans: true,
              isPosted: isPosted);
        }
      } else {
        if (service != null) {
          var serviceResult = await service.beforeUpdate(context, data);

          if (!serviceResult.success) {
            return serviceResult;
          }
          if (!serviceResult.skip && serviceResult.success) {
            data = serviceResult.data;
            // result = await EntityService._updateEntity(
            //     entity, serviceResult.data,
            //     localOnly: localOnly, withTrans: true);
          }
        }
        result = await EntityService._updateEntity(entity, data,
            localOnly: localOnly,
            withTrans: true,
            forceOnline: forceOnline,
            isPosted: isPosted,
            entitySaveConfirm: entitySaveConfirm);
      }

      if (!result.success) return result;

      if (ConnectionService.isOfflineMode && !forceOnline) {
        for (var prop in entity?.properties ?? []) {
          if (data[prop.dbName] == null) continue;
          if ((isAddNew ? prop.isAdd : prop.isEdit) &&
              prop.propertyType == PropertyTypes.ManyToMany) {
            var property = prop as PropertyManyToMany;
            //var dataList = data
            if (property.useArrayStringValue) {
              var newVal = data[prop.dbName] is List<String>
                  ? data[prop.dbName] as List<String>
                  : (data[prop.dbName] as List<dynamic>).cast<String>();
              data[prop.dbName + "Value"] = newVal.join(',');
              continue;
            }
            var manyToManyUpdateResult =
                await EntityService._updateManyToManyEntity(
                    prop, data, data['id']);
            if (!manyToManyUpdateResult.success) {
              return manyToManyUpdateResult;
            }
          } else if ((isAddNew ? prop.isAdd : prop.isEdit) &&
              prop.propertyType == PropertyTypes.OtherEntityMany) {
            var newItems = data[prop.dbName];
            if (newItems == null) continue;
            var property = prop as PropertyOtherEntityMany;

            var otherEntity = EntityService.getEntity(property.entityName);
            // read the old rows

            var otherPropertyType = "${entity?.dbName}Id";

            List<Map<String, dynamic>> oldRows = isAddNew
                ? []
                : (await getData(otherEntity, filter: [
                      EntityFilter(
                          value: data['id'],
                          operatorId: "E",
                          propertyName: "${entity?.dbName}Id")
                    ]))
                        .data ??
                    [];

            // update the updated
            for (var newItem in newItems) {
              // set the relation key
              newItem[otherPropertyType] = data['id'];
              bool isExist = oldRows.any((e) => e['id'] == newItem['id']);
              // add or update
              OperationResult updateResult = await EntityService.saveEntity(
                  otherEntity, newItem, null, !isExist,
                  withTrans: true);
              if (!updateResult.success) {
                return updateResult;
              }
            }

            if (oldRows.length > 0) {
              // delete the removed
              var removedRows = oldRows
                  .where((r) => !newItems.any((e) => e['id'] == r['id']))
                  .map((r) => EntityFilter(
                      value: r['id'], operatorId: 'E', propertyName: 'id'))
                  .toList();

              if (removedRows.length > 0) {
                var removeResult = await DatabaseService.removeRange(
                    otherEntity?.dbName ?? '', removedRows,
                    withTrans: true, divider: 'or');

                if (!removeResult)
                  return OperationResult(
                      success: false,
                      message: 'unable to delete other entity ${prop.dbName}');
              }
            }
            // remove it from the model because it's already saved
            data[prop.dbName] = null;
          }
        } // end loop in each prop
      } // end if offline

      // after add or update
      if (service != null) {
        OperationResultEntityTransaction serviceResult;
        if (isAddNew)
          serviceResult = await service.afterAdd(context, data);
        else
          serviceResult = await service.afterUpdate(context, data);

        if (!serviceResult.success) {
          return serviceResult;
        }
      }
      // print(
      //     '||||||||||||||||||||| trans ${entity?.dbName} ||||||||||||||||||||||||||||||||||||');
      // print(DatabaseService.trans?.map((t) =>
      //     '${t.table} : id=${t.data != null ? t.data!['id'] : null} : ${t.query}'));
      var mainResult = OperationResult(
          success: withTrans || ConnectionService.isOnlineMode
              ? true
              : await DatabaseService.execTrans(),
          data: result.data);
      if (mainResult.success && TecfySyncService.isOfflineMode) {
        TecfySyncService.pushNow();
      }
      return mainResult;
    } catch (err, tack) {
      print('XXXXXXXXXXXXXXX Error Save Entity: $err');
      print(tack);
      print(data);
      return OperationResult(success: false, message: 'Error: $err');
    }
  }

  /// custom Encode
  static dynamic customEncode(dynamic item) {
    if (item is DateTime) {
      return intl.DateFormat("yyyy-MM-ddTHH:mm:ss").format(item);
    }
    return item;
  }

  /// lower First char
  static String lowerFirst(String? s) {
    if (s == null || s.isEmpty) return '';
    return '${s[0].toLowerCase()}${s.substring(1)}';
  }

  /// upper First char
  static String upperFirst(String? s) {
    if (s == null || s.isEmpty) return '';
    return '${s[0].toUpperCase()}${s.substring(1)}';
  }

  /// get bool value from dynamic
  static bool getBool(dynamic data) {
    if (data == null) return false;
    if (data is bool) return (data == true);
    if (data is num) return (data == 1);
    return data.toString() == '1' || data.toString() == 'true';
  }

  static Future<OperationResult> _updateManyToManyEntity(
      PropertyManyToMany prop, Map<String, dynamic> model, String parentId,
      {bool withTrans = false}) async {
    print('--------------- start save many to many ${prop.dbName}');
    try {
      var middleEntityType = EntityService.getEntity(prop.middleEntityName);
      if (middleEntityType == null) {
        return OperationResult(
            success: false, message: 'Entity cannot be null');
      }
      var oldItems = await getData(middleEntityType, filter: [
        EntityFilter(
            value: parentId,
            operatorId: "E",
            propertyName: prop.middlePropertyName)
      ]);

      var middleOtherPropertyNameProp = prop.middleOtherPropertyName;
      var oldIds =
          oldItems.data?.map((s) => s[middleOtherPropertyNameProp]).toList() ??
              [];

      var propModel = prop.dbName;

      if (model[propModel] == null) {
        return OperationResult(success: true);
      }

      var newItems = model[propModel] is List<Map<String, dynamic>>
          ? model[propModel] as List<Map<String, dynamic>>
          : (model[propModel] as List<dynamic>).cast<Map<String, dynamic>>();
      // ((IEnumerable)propModel.GetValue(model)).OfType<object>();

      model[propModel] = null;

      //Type newItemType = null;
      var newItemIdProp;
      if (newItems.length > 0) {
        //newItemType = newItems.First().GetType();
        newItemIdProp = prop.middleOtherPropertyName; //Updated "Id"
      }
      List<dynamic> ids;
      if (prop.isWithData) {
        ids = newItems.map((s) => s[newItemIdProp]).toList();
      } else {
        ids = newItems.toList();
      }

      Entity? middlePropertyStructure;
      if (prop.isWithData)
        middlePropertyStructure =
            EntityService.getEntity(prop.middleEntityName);

      // add the new selected items
      var newIds = ids.where((s) => !oldIds.contains(s) && s != null);
      if (newIds.length > 0) {
        for (var id in newIds) {
          dynamic newItem;
          if (prop.isWithData) {
            newItem = newItems.firstWhere((i) => i[newItemIdProp] == id);
          }

          dynamic middleEntity = {};

          var entity = middleEntityType;

          if (prop.isWithData) {
            for (var tempPropStructure in entity.properties.where((p) =>
                p.isEdit && p.propertyType != PropertyTypes.ManyToMany)) {
              var newValue = newItem[tempPropStructure.dbName];
              middleEntity[tempPropStructure.dbName] = newValue;
            }
          }

          middleEntity[prop.middlePropertyName] = parentId;
          middleEntity[prop.middleOtherPropertyName] = id;

          var newObject = await EntityService.saveEntity(
              middleEntityType, middleEntity, null, true,
              withTrans: true);
          if (!newObject.success) return newObject;
        }
      }

      // delete the removed items

      var removedIds = oldIds.where((s) => !ids.contains(s));
      if (removedIds.length > 0) {
        //var queryToRemove = EntityService.getEntity(middleEntityType, parentId, "E", prop.middlePropertyName);
        var filter = EntityFilter(
            value: parentId,
            operatorId: "E",
            propertyName: prop.middlePropertyName);
        for (var i in removedIds) {
          var filter2 = EntityFilter(
              value: i,
              operatorId: "E",
              propertyName: prop.middleOtherPropertyName);
          await DatabaseService.removeRange(prop.entityName, [filter, filter2],
              withTrans: withTrans);
        }
      }

      // update the changes
      if (prop.isWithData) {
        var needUpdateItems = oldItems.data
                ?.where((s) => ids.contains(s[prop.middleOtherPropertyName])) ??
            [];

        for (var needUpdateItem in needUpdateItems) {
          dynamic oldItem;
          var id = needUpdateItem[prop.middleOtherPropertyName];

          oldItem = newItems.firstWhere((s) => s[newItemIdProp] == id);

          for (var p in middlePropertyStructure?.properties ?? []) {
            if (p.propertyType == PropertyTypes.ManyToMany) {
              var resultAddMany = await _updateManyToManyEntity(
                  p as PropertyManyToMany, oldItem, needUpdateItem['id']);
              if (!resultAddMany.success) {
                return resultAddMany;
              }
            } else {
              needUpdateItem[p.dbName] = oldItem[p.dbName];
            }
          }
        }
      }

      return OperationResult(success: true);
    } catch (err, trace) {
      print('XXXXXXXXXXXXXXXXXXX ManytoManyError');
      print(err);
      print(trace);
      return new OperationResult(
          success: false, message: 'Unable to update ${prop.dbName}');
    }
  }

  static List<EAction> getActions(Entity entity, Map<String, dynamic>? d) {
    List<EAction> a = entity.actions
        .map((a) => EAction(
            dbName: a.dbName,
            icon: Icon(_getIconFromString(a.icon)),
            iconData: _getIconFromString(a.icon),
            nameAr: a.nameAr,
            nameEn: a.nameEn,
            serverSideAction: a))
        .where((act) =>
            d == null ||
            (act.serverSideAction?.isVisible(d, entity.properties) ?? true))
        .toList();

    var entityActionsService =
        TecfyCoreApp.entityActionsServices[entity.dbName.toLowerCase()];
    if (entityActionsService == null) return a;
    a.addAll(entityActionsService.getActions(d).data ?? []);
    return a;
  }

  static Future onAction(Entity entity, EAction action,
      Map<String, dynamic> data, BuildContext context) async {
    if (action.serverSideAction != null) {
      await _onServersideActionClicked(
          entity, action.serverSideAction!, data, context);
    } else {
      var entityActionsService =
          TecfyCoreApp.entityActionsServices[entity.dbName.toLowerCase()];
      await entityActionsService?.execAction(action.dbName, data, context);
    }
  }

  static Future _onServersideActionClicked(Entity entity, EntityAction action,
      Map<String, dynamic> data, BuildContext context) async {
    var form = Map<String, dynamic>();
    if (action.overwriteProperties != null &&
        action.overwriteProperties!.isNotEmpty) {
      var props = action.overwriteProperties!.split(',');
      for (var p in props) {
        form[EntityService.lowerFirst(p.split(':').last)] =
            data[EntityService.lowerFirst(p.split(':').first)];
      }
    }
    var result = await AppNavigator.showAppDialogEx(
          context,
          () => Padding(
            padding: EdgeInsets.all(10),
            child: Column(
              children: <Widget>[
                Chip(label: Text('${action.name}')),
                Expanded(
                    child: SingleChildScrollView(
                        child: Column(
                  children: action.actionFormProp
                      .map((p) => PorpertyComponent(
                          data: form,
                          property: p,
                          onChange: (v) => form[p.dbName] = v))
                      .toList(),
                ))),
                SizedBox(height: 10),
                AppButtonPost(
                    label: action.name,
                    onPressed: () async {
                      Navigator.pop(context, true);
                    }),
                SizedBox(height: 10),
              ],
            ),
          ),
        ) ??
        false;
    // print(result);
    // print(form);
    if (result) {
      for (var prop in action.actionFormProp) {
        if (prop.isRequired && form[prop.dbName] == null) {
          AppNavigator.showMessage(
              context, 'Property required : ${prop.name}', MessageType.warning);
          return;
        }
      }
      AppNavigator.showLoading(context);
      form['id'] = data['id'];
      var postResult = await TecfyBasicApp.apiService.httpPost(
          'Entity/${entity.dbName}/action/${action.dbName}', form, true);
      AppNavigator.pop(context);
      AppNavigator.showMessage(
          context,
          postResult.success ? 'Done' : postResult.message ?? 'Unknown Error!',
          postResult.success ? MessageType.success : MessageType.error);
    }
  }

  static IconData _getIconFromString(String? icon,
      [IconData defaultIcon = Icons.import_export_rounded]) {
    if (icon == null || icon.isEmpty) return defaultIcon;
    switch (icon) {
      case 'cancel':
        return Icons.cancel;
      case 'delete':
        return Icons.delete;
      case 'print':
        return Icons.print;
      case 'approve':
        return Icons.done;
      case 'done':
        return Icons.done;
      case 'done_all':
        return Icons.done_all;
      case 'schedule':
        return Icons.schedule;
      case 'time':
        return Icons.schedule;
      case 'alarm':
        return Icons.access_alarm;
      case 'send':
        return Icons.send;
      case 'reset':
        return Icons.reset_tv;
      case 'refresh':
        return Icons.refresh;
      case 'money':
        return Icons.attach_money;
      case 'pay':
        return Icons.payment;
      case 'refund':
        return Icons.money_off;
      case 'start':
        return Icons.play_circle_outline_sharp;
      case 'play':
        return Icons.play_circle_outline_sharp;
      case 'stop':
        return Icons.stop_circle_outlined;
      default:
        break;
    }
    return defaultIcon;
  }
}
